/* eslint camelcase: "off" */
const fs = require('fs');
const path = require('path');
const childProcess = require('child_process');
const FfmpegMediaFile = require('../FfmpegMediaFile/FfmpegMediaFile');
const FfmpegStream = require('../FfmpegStream/FfmpegStream');
const { InputStreamFactory } = require('../FfmpegStreamFactory/FfmpegStreamFactory');
const FfmpegParsedData = require('../FfmpegParsedData/FfmpegParsedData');
const FfmpegValidators = require('../FfmpegValidators/FfmpegValidators');
const { FfmpegUnsupportedError } = require('../FfmpegErrors/FfmpegErrors');
const { getSupportedVideoDecoders, getSupportedAudioDecoders } = require('../FfmpegProgram/FfmpegProgram');
require('dotenv').config();

const { env: { FFPROBE_BIN } } = process;

class FfmpegInput extends FfmpegMediaFile {
  /**
   * @constructor
   * @param {string|Object} newInputFile path to an input file, or object containing file metadata.
   */
  constructor(newInputFile) {
    super();
    if (typeof newInputFile === 'string') {
      this.setFile(newInputFile);
      this.parsedData = new FfmpegParsedData();
      this.streams = [];
    } else if (newInputFile instanceof FfmpegInput) {
      this.setFile(newInputFile.getFullFilePath());
      this.parsedData = new FfmpegParsedData(newInputFile.parsedData);
      this.streams = this.parsedData.streams.map(stream => InputStreamFactory(stream));
    } else if (typeof newInputFile === 'object') {
      this.setFile(`${newInputFile.path}/${newInputFile.file}`);
      this.parsedData = new FfmpegParsedData(newInputFile.parsedData);
      this.streams = this.parsedData.streams.map(stream => InputStreamFactory(stream));
    } else {
      throw new TypeError('new FfmpegInput() requires a string or object as its parameter');
    }
  }

  /* Begin methods that map directly to ffmpeg options */

  acodec(codec) {
    super.acodec(codec, getSupportedAudioDecoders());
    return this;
  }

  /**
   * Set number of times input stream shall be looped. Loop 0 means no loop, loop -1 means
   * infinite loop.
   *
   * @param {Number} number Loop count. Requires integer, floats will be truncated
   */
  stream_loop(number) {
    const asInt = FfmpegValidators.toInteger(number);

    if (asInt < -1) {
      throw new TypeError('stream_loop() requires an integer argument >= -1');
    }

    return this.addOption('stream_loop', asInt);
  }

  /**
   * Like the -ss option but relative to the "end of file". That is negative values are earlier
   * in the file, 0 is at EOF.
   */
  sseof(position) {
    const time = new FfmpegValidators.FfmpegTime(position);
    if (time.isZero() || time.isNegative()) {
      return this.addOption('sseof', time.get());
    }

    throw new TypeError('sseof() requires 0, a negative value, or a string in the format -[HH:]mm:ss[.MM]');
  }

  /**
   * Set the input time offset. The offset is added to the timestamps of the input files.
   * Specifying a positive offset means that the corresponding streams are delayed by the time
   * duration specified in offset.
   *
   * @param {number} offset must be a time duration specification, see the Time duration section
   * in the ffmpeg-utils manual.
   */
  itsoffset(offset) {
    const time = new FfmpegValidators.FfmpegTime(offset);
    if (time.isValid()) {
      this.addOption('itsoffset', offset);
      return this;
    }

    throw new TypeError('sseof() requires an integer or a string in the format [-][HH:]mm:ss[.MM]');
  }

  /**
   * Read input at native frame rate. Mainly used to simulate a grab device, or live input stream
   * (e.g. when reading from a file). Should not be used with actual grab devices or live input
   * streams (where it can cause packet loss). By default ffmpeg attempts to read the input(s) as
   * fast as possible. This option will slow down the reading of the input(s) to the native frame
   * rate of the input(s). It is useful for real-time output (e.g. live streaming).
   */
  re() {
    return this.addOption('re');
  }

  /**
   * Set the maximum demux-decode delay.
   */
  muxdelay(seconds) {
    const number = FfmpegValidators.toFloat(seconds);
    return this.addOption('muxdelay', number);
  }

  /**
   * Set the initial demux-decode delay.
   */
  muxpreload(seconds) {
    const number = FfmpegValidators.toFloat(seconds);
    return this.addOption('muxpreload', number);
  }

  /**
   * This option enables or disables accurate seeking in input files with the -ss option. It is
   * enabled by default, so seeking is accurate when transcoding. Use -noaccurate_seek to disable
   * it, which may be useful e.g. when copying some streams and transcoding the others.
   */
  accurate_seek() {
    this.removeOption('noaccurate_seek');
    return this.addOption('accurate_seek');
  }

  noaccurate_seek() {
    this.removeOption('accurate_seek');
    return this.addOption('noaccurate_seek');
  }

  /**
   * This option enables or disables seeking by timestamp in input files with the -ss option. It
   * is disabled by default. If enabled, the argument to the -ss option is considered an actual
   * timestamp, and is not offset by the start time of the file. This matters only for files which
   * do not start from timestamp 0, such as transport streams.
   */
  seek_timestamp() {
    return this.addOption('-seek_timestamp');
  }

  /**
   * This option sets the maximum number of queued packets when reading from the file or device.
   * With low latency / high rate live streams, packets may be discarded if they are not read in
   * a timely manner; raising this value can avoid it.
   */
  thread_queue_size(size) {
    const number = FfmpegValidators.toFloat(size);
    return this.addOption('-thread_queue_size', number);
  }

  /**
   * Allows discarding specific streams or frames from streams. Any input stream can be fully
   * discarded, using value all whereas selective discarding of frames from a stream occurs at
   * the demuxer and is not supported by all demuxers.
   */
  discard(setting) {
    const validArguments = ['none', 'default', 'noref', 'bidir', 'nokey', 'all'];

    if (validArguments.includes(setting)) {
      return this.addOption('-discard', setting);
    }

    throw new FfmpegUnsupportedError('Valid arguments for discard() are "none", "default", "noref", "bidir", "nokey", "all"');
  }

  vcodec(codec) {
    super.vcodec(codec, getSupportedVideoDecoders());
    return this;
  }

  /* End methods that map directly to Ffmpeg options */

  filter(filter) {
    const fieldName = Object.keys(filter)[0];
    return this.streams.filter(stream => filter[fieldName] === stream[fieldName]);
  }

  videoStreams() {
    return this.filter({ codec_type: 'video' });
  }

  audioStreams() {
    return this.filter({ codec_type: 'audio' });
  }

  subtitleStreams() {
    return this.filter({ codec_type: 'subtitle' });
  }

  dataStreams() {
    return this.filter({ codec_type: 'data' });
  }

  largestDimensions() {
    const video = this.videoStreams();

    if (video.length < 1) {
      return null;
    }

    if (video.length === 1) {
      return video[0];
    }

    return video.reduce((winner, current) => {
      const winnerSize = winner.coded_width * winner.coded_height;
      const currentSize = current.coded_width * current.coded_height;
      return currentSize > winnerSize ? current : winner;
    });
  }

  mostChannels() {
    const audio = this.audioStreams();

    if (audio.length < 1) {
      return undefined;
    }

    if (audio.length === 1) {
      return audio[0];
    }

    return audio.reduce((winner, current) => (
      current.channels > winner.channels ? current : winner
    ));
  }

  setFile(newInputFile) {
    if (typeof newInputFile !== 'string') {
      throw new TypeError('input file must be a string');
    }

    fs.accessSync(newInputFile, fs.constants.R_OK);

    this.file = path.basename(newInputFile);
    this.path = path.dirname(newInputFile);
    delete this.format;
    delete this.streams;

    return this;
  }

  /**
   * @returns {string} The filename of the file, without the path.
   */
  getFileName() {
    return this.file;
  }

  /**
   * @returns {string} The path of the file, without a trailing separator.
   */
  getFilePath() {
    return this.path;
  }

  getFullFilePath() {
    return `${this.path}/${this.file}`;
  }

  /**
   * Synchronously reads the input file's metadata using Ffprobe, and stores it.
   * @returns {FfmpegInput} *this*
   */
  read() {
    const ffprobeArguments = ['-i', this.getFullFilePath(), '-v', 'quiet', '-show_format', '-show_streams', '-print_format', 'json=compact=1'];

    // Matroska files will not return a frame count without this switch (which also takes longer)
    if (/\.mkv$/i.test(this.getFileName())) {
      ffprobeArguments.push('-count_frames');
    }

    const json = JSON.parse(childProcess.execFileSync(FFPROBE_BIN, ffprobeArguments));

    this.saveParsedData(json);

    return this.getParsedData();
  }

  /**
   * Asynchronously reads the input file's metadata using Ffprobe, and stores it.
   * @returns {Promise|Error} Resolves with *this*, rejects with Error.
   */
  readAsync() {
    // Matroska files will not return a frame count without this switch (which also takes longer)
    const needsDeepFrameCount = /\.mkv$/i.test(this.getFileName());

    const ffprobeArguments = ['-i', this.getFullFilePath(), '-v', 'quiet', '-show_format', '-show_streams', '-print_format', 'json=compact=1'];
    if (needsDeepFrameCount) ffprobeArguments.push('-count_frames');

    const ffprobeChild = childProcess.execFile(FFPROBE_BIN, ffprobeArguments);

    let output = '';
    ffprobeChild.stdout.on('data', (data) => { output += data; });

    return new Promise((resolve, reject) => {
      ffprobeChild.on('error', error => reject(error));

      ffprobeChild.stdout.once('close', () => {
        const json = JSON.parse(output);

        if (needsDeepFrameCount) {
          this.streams.forEach(stream => stream.nb_frames = stream.nb_read_frames); // eslint-disable-line
        }

        this.parsedData = new FfmpegParsedData(json);
        this.streams = this.parsedData.streams.map(stream => InputStreamFactory(stream));
        resolve(this.getParsedData());
      });
    });
  }

  saveParsedData(data) {
    this.parsedData = new FfmpegParsedData(data);
    this.streams = this.parsedData.streams.map(stream => new FfmpegStream(stream));
    return this;
  }

  /**
   * Returns the FfmpegInput's properties as an object, optionally converted to a JSON string
   *
   * @param {Boolean}        [stringify=false]  Return a JSON string rather than an object
   *
   * @returns {Object|String} This object, as an object, or a string if stringify=true
   */
  getParsedData(stringify = false) {
    return this.parsedData.get(stringify);
  }

  getCommandLine() {
    const commandLine = super.getCommandLine();
    this.streams.reduce((acc, curr) => acc.concat(curr.getCommandLine()), []);
    return commandLine.concat(['-i', this.getFullFilePath()]);
  }
}

module.exports = FfmpegInput;
