/* eslint-env mocha */
/* eslint-disable no-unused-expressions */
const { expect } = require('chai');
const FfmpegInput = require('./FfmpegInput');
const FfmpegStream = require('../FfmpegStream/FfmpegStream');
const { inputFileMock, inputQuicktimePath } = require('../../test/fixtures');
require('dotenv').config();

describe('FfmpegInput', () => {
  describe('Unit Tests', () => {
    describe('Constructor', () => {
      it('accepts a string parameter', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(test.file).to.equal('IMG_0246.MOV');
        expect(test.path).to.equal(`${process.cwd()}/test`);
      });

      it('accepts an object parameter', () => {
        const test = new FfmpegInput(inputFileMock);
        expect(test.file).to.equal('IMG_0246.MOV');
        expect(test.path).to.equal(`${process.cwd()}/test`);
      });

      it('accepts FfmpegInput parameter', () => {
        const firstCopy = new FfmpegInput(inputQuicktimePath);
        const secondCopy = new FfmpegInput(firstCopy);
        expect(firstCopy).to.deep.equal(secondCopy, 'path property was not copied from input');
        expect(firstCopy).to.not.equal(secondCopy, 'format property was shallow copied');
      });

      it('throws without a parameter', () => {
        expect(() => new FfmpegInput()).to.throw(TypeError);
      });

      it('throws with a number parameter', () => {
        expect(() => new FfmpegInput(42)).to.throw(TypeError);
      });

      it('throws with a nonexistent file path string parameter', () => {
        expect(() => new FfmpegInput('fhjdfkslkjl')).to.throw(Error);
      });
    });

    describe('setFile()', () => {
      it('sets an input file with a string argument', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.setFile(inputQuicktimePath);
        expect(test.path).to.equal(`${process.cwd()}/test`);
        expect(test.file).to.equal('IMG_0246.MOV');
      });

      it('throws with a number argument', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.setFile(2)).to.throw();
      });

      it('throws with an object argument', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.setFile({})).to.throw();
      });
    });

    describe('getFileName()', () => {
      it('Returns the filename', () => {
        const input = new FfmpegInput(inputQuicktimePath);
        expect(input.getFileName()).to.equal('IMG_0246.MOV');
      });
    });

    describe('getFilePath()', () => {
      it('Returns the file path', () => {
        const input = new FfmpegInput(inputQuicktimePath);
        expect(input.getFilePath()).to.equal(`${process.cwd()}/test`);
      });
    });

    describe('getFullFilePath()', () => {
      it('Returns the file path and file name', () => {
        const input = new FfmpegInput(inputQuicktimePath);
        expect(input.getFullFilePath()).to.equal(`${process.cwd()}/test/IMG_0246.MOV`);
      });
    });

    describe('read()', () => {
      it('reads and saves a file\'s metadata', () => {
        const input = new FfmpegInput(inputQuicktimePath);
        expect(input.parsedData.format, 'format was set before reading file').to.deep.equal({});
        expect(input.parsedData.streams, 'streams was set before reading file').to.deep.equal([]);
        input.read();
        expect(input.parsedData.format).to.deep.equal(inputFileMock.parsedData.format);
        expect(input.parsedData.streams).to.deep.equal(inputFileMock.parsedData.streams);
      });

      it('throws with an invalid media file', () => {
        const test = new FfmpegInput(`${process.cwd()}/test/fixtures.js`);

        expect(() => test.read()).to.throw();
      });

      it('returns parsed data', () => {
        const input = new FfmpegInput(inputQuicktimePath);
        const test = input.read();
        expect(test.format).to.deep.equal(inputFileMock.parsedData.format);
        expect(test.streams).to.deep.equal(inputFileMock.parsedData.streams);
      });
    });

    describe('readAsync()', () => {
      it('reads and saves a file\'s metadata', async () => {
        const input = new FfmpegInput(inputQuicktimePath);
        expect(input.parsedData.format, 'format was set before reading file').to.deep.equal({});
        expect(input.parsedData.streams, 'streams was set before reading file').to.deep.equal([]);
        const test = await input.readAsync();
        expect(test.format).to.deep.equal(inputFileMock.parsedData.format);
        expect(test.streams).to.deep.equal(inputFileMock.parsedData.streams);
      });

      it('throws with an invalid media file', () => {
        const test = new FfmpegInput(`${process.cwd()}/test/fixtures.js`);

        expect(test.readAsync).to.throw();
      });
    });
  });


  describe('StreamFilters', () => {
    describe('filter()', () => {
      it('returns an array', () => {
        const test = new FfmpegInput(inputFileMock).filter({ codec_name: 'h264' });
        expect(test).to.be.an('array');
      });

      it('returns a stream with "codec_name": "h264"', () => {
        const input = new FfmpegInput(inputFileMock);
        const test = input.filter({ codec_name: 'h264' });
        expect(test[0]).to.deep.equal(input.streams[1]);
      });

      it('returns an empty Array when nothing matches filter', () => {
        const test = new FfmpegInput(inputFileMock).filter({ codec_type: 'foo' });
        expect(test).to.deep.equal([]);
      });
    });

    describe('videoStreams()', () => {
      it('returns an array', () => {
        const test = new FfmpegInput(inputFileMock).videoStreams();
        expect(test).to.be.an('array');
      });

      it('returned collection contains a single stream for this test fixture', () => {
        const test = new FfmpegInput(inputFileMock).videoStreams();
        expect(test.length).to.equal(1);
      });

      it('returned collection contains only instances of FfmpegStream', () => {
        const test = new FfmpegInput(inputFileMock).videoStreams();
        test.forEach((stream, index) => {
          expect(stream, `Element at index ${index} had is not an FfmpegStream`).to.be.an.instanceOf(FfmpegStream);
        });
      });

      it('returned collection contains only video streams', () => {
        const test = new FfmpegInput(inputFileMock).videoStreams();
        test.forEach((stream, index) => {
          expect(stream.codec_type, `Stream at index ${index} had type ${stream.codec_type}`).to.equal('video');
        });
      });
    });
  });

  describe('Ffmpeg options methods', () => {
    describe('acodec()', () => {
      it('adds c:a to options', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.acodec('aac');
        expect(test.ffmpegSettings.options).to.deep.equal({ 'c:a': 'aac' });
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        const test2 = test.acodec('aac');
        expect(test2).to.equal(test);
      });

      it('throws when passed an unsupported codec', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.acodec('jkfshkkj')).to.throw();
      });
    });

    describe('vcodec()', () => {
      it('adds c:v to options', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.vcodec('hevc');
        expect(test.ffmpegSettings.options).to.deep.equal({ 'c:v': 'hevc' });
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        const test2 = test.vcodec('hevc');
        expect(test2).to.equal(test);
      });

      it('throws when passed an unsupported codec', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.vcodec('jkfshkkj')).to.throw();
      });
    });

    describe('stream_loop()', () => {
      it('adds stream_loop to options with a positive argument', () => {
        const test5 = new FfmpegInput(inputQuicktimePath);
        test5.stream_loop(5);
        expect(test5.ffmpegSettings.options).to.deep.equal({ stream_loop: 5 });

        const test326 = new FfmpegInput(inputQuicktimePath);
        test326.stream_loop(326);
        expect(test326.ffmpegSettings.options).to.deep.equal({ stream_loop: 326 });
      });

      it('adds stream_loop to options with 0', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.stream_loop(0);
        expect(test.ffmpegSettings.options).to.deep.equal({ stream_loop: 0 });
      });

      it('adds stream_loop to options with -1', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.stream_loop(-1);
        expect(test.ffmpegSettings.options).to.deep.equal({ stream_loop: -1 });
      });

      it('throws when called with a non-number', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.stream_loop('fdsf')).to.throw();
      });

      it('throws when called with a number < -1', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.stream_loop('-2')).to.throw();
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        const test2 = test.stream_loop(5);
        expect(test2).to.equal(test);
      });
    });

    describe('sseof()', () => {
      it('adds sseof to options with a negative argument', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.sseof(-60);
        expect(test.ffmpegSettings.options).to.deep.equal({ sseof: -60 });
      });

      it('throws when called with an invalid time string', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.sseof('1:5.6.7')).to.throw();
      });

      it('throws with a positive argument', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.sseof(60)).to.throw();
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        const test2 = test.sseof(-60);
        expect(test2).to.equal(test);
      });
    });

    describe('itsoffset()', () => {
      it('adds itsoffset to options with an integer argument', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.itsoffset(-60);
        expect(test.ffmpegSettings.options).to.deep.equal({ itsoffset: -60 });
      });

      it('adds itsoffset to options with a time string argument', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.itsoffset('01:25');
        expect(test.ffmpegSettings.options).to.deep.equal({ itsoffset: '01:25' });
      });

      it('throws when called with an invalid time string', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.itsoffset('1:5.6.7')).to.throw();
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        const test2 = test.itsoffset(60);
        expect(test2).to.equal(test);
      });
    });

    describe('re()', () => {
      it('adds re to options', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.re();
        expect(test.ffmpegSettings.options).to.deep.equal({ re: null });
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        const test2 = test.re();
        expect(test2).to.equal(test);
      });
    });

    describe('muxdelay()', () => {
      it('adds muxdelay to options with a number argument', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.muxdelay(5);
        expect(test.ffmpegSettings.options).to.deep.equal({ muxdelay: 5 });
      });

      it('throws when called with a non-number', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.muxdelay('fdsf')).to.throw();
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        const test2 = test.stream_loop(5);
        expect(test2).to.equal(test);
      });
    });

    describe('muxpreload()', () => {
      it('adds muxpreload to options with a number argument', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.muxpreload(5);
        expect(test.ffmpegSettings.options).to.deep.equal({ muxpreload: 5 });
      });

      it('throws when called with a non-number', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.muxpreload('fdsf')).to.throw();
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        const test2 = test.muxpreload(5);
        expect(test2).to.equal(test);
      });
    });

    describe('accurate_seek()', () => {
      it('adds re to options', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.accurate_seek();
        expect(test.ffmpegSettings.options).to.deep.equal({ accurate_seek: null });
      });

      it('removes noaccurateseek if set', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.noaccurate_seek();
        test.accurate_seek();
        expect(test.ffmpegSettings.options).to.deep.equal({ accurate_seek: null });
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        const test2 = test.accurate_seek();
        expect(test2).to.equal(test);
      });
    });

    describe('noaccurate_seek()', () => {
      it('adds re to options', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.noaccurate_seek();
        expect(test.ffmpegSettings.options).to.deep.equal({ noaccurate_seek: null });
      });

      it('removes accurateseek if set', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.accurate_seek();
        test.noaccurate_seek();
        expect(test.ffmpegSettings.options).to.deep.equal({ noaccurate_seek: null });
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        const test2 = test.noaccurate_seek();
        expect(test2).to.equal(test);
      });
    });

    describe('seek_timestamp()', () => {
      it('adds seek_timestamp to options', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.seek_timestamp();
        expect(test.ffmpegSettings.options).to.deep.equal({ seek_timestamp: null });
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        const test2 = test.seek_timestamp();
        expect(test2).to.equal(test);
      });
    });

    describe('thread_queue_size()', () => {
      it('adds thread_queue_size to options with a number argument', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.thread_queue_size(5);
        expect(test.ffmpegSettings.options).to.deep.equal({ thread_queue_size: 5 });
      });

      it('throws when called with a non-number', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.thread_queue_size('fdsf')).to.throw();
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        const test2 = test.thread_queue_size(5);
        expect(test2).to.equal(test);
      });
    });

    describe('discard()', () => {
      it('adds discard to options with a valid', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        test.discard('nokey');
        expect(test.ffmpegSettings.options).to.deep.equal({ discard: 'nokey' });
      });

      it('throws when called with a non-number', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.discard('fdsf')).to.throw();
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        const test2 = test.discard('nokey');
        expect(test2).to.equal(test);
      });
    });
  });

  describe('Inheritance/Composition tests', () => {
    describe('addOption()', () => {
      it('adds an argument', () => {
        const test = new FfmpegInput(inputFileMock);
        test.addOption('ss', 60);
        expect(test.hasOption('ss')).to.be.true;
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputFileMock);
        const test2 = test.addOption('ss', 60);
        expect(test).to.equal(test2);
      });
    });

    describe('removeOption()', () => {
      it('removes an option', () => {
        const test = new FfmpegInput(inputFileMock);
        test.addOption('ss', 60);
        expect(test.hasOption('ss')).to.be.true;
        test.removeOption('ss');
        expect(test.hasOption('ss')).to.be.false;
      });

      it('returns this', () => {
        const test = new FfmpegInput(inputFileMock);
        const test2 = test.removeOption();
        expect(test).to.equal(test2);
      });
    });

    describe('setFile()', () => {
      it('returns this', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(test.file).to.equal('IMG_0246.MOV', 'Initial input file could not be set');
        const copy = test.setFile(inputQuicktimePath.replace('6', '7'));
        expect(test).to.equal(copy);
      });

      it('throws TypeError when called without a parameter', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(test.setFile).to.throw(TypeError);
      });


      it('throws when Error called a non-existing file parameter', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.setFile('dfsdfds')).to.throw(Error);
      });

      it('throws TypeError when called with a non-string parameter', () => {
        const test = new FfmpegInput(inputQuicktimePath);
        expect(() => test.setFile(42)).to.throw(TypeError);
      });
    });

    describe('getFileName()', () => {
      it('returns a string', () => {
        const file = new FfmpegInput(inputFileMock).getFileName();
        expect(file).to.be.a('string');
      });

      it('returns only the base file name', () => {
        const file = new FfmpegInput(inputFileMock).getFileName();
        expect(file).to.equal('IMG_0246.MOV');
      });
    });

    describe('getFilePath()', () => {
      it('returns a string', () => {
        const file = new FfmpegInput(inputFileMock).getFilePath();
        expect(file).to.be.a('string');
      });

      it('returns the full file path', () => {
        const file = new FfmpegInput(inputFileMock).getFilePath();
        expect(file).to.equal(`${process.cwd()}/test`);
      });
    });

    describe('getFullFilePath()', () => {
      it('Returns the file path and file name', () => {
        const input = new FfmpegInput(inputFileMock);
        expect(input.getFullFilePath()).to.equal(`${process.cwd()}/test/IMG_0246.MOV`);
      });
    });

    describe('getParsedData', () => {
      it('returns an object when called with no arguments', () => {
        const input = new FfmpegInput(inputFileMock);
        expect(input.getParsedData()).to.be.an('object');
      });

      it('returns a string when called with true', () => {
        const input = new FfmpegInput(inputFileMock);
        expect(input.getParsedData(true)).to.be.a('string');
      });

      it('returns the format and streams properties', () => {
        const input = new FfmpegInput(inputFileMock);
        expect(input.getParsedData()).to.be.deep.equal(inputFileMock.parsedData);
      });
    });
  });

  describe('Integration Tests', () => {
    describe('getCommandLine()', () => {
      it('returns an array', () => {
        const test = new FfmpegInput(inputFileMock).getCommandLine();
        expect(test).to.be.an('array');
      });

      it('returns only input argument for an input with no added arguments', () => {
        const testArguments = new FfmpegInput(inputFileMock).getCommandLine();
        expect(testArguments).to.deep.equal(['-i', inputQuicktimePath]);
      });

      it('returns input settings and input argument for input with one added argument', () => {
        const input = new FfmpegInput(inputFileMock);
        input.addOption('ss', 60);
        const testArguments = input.getCommandLine();
        expect(testArguments).to.deep.equal(['-ss', 60, '-i', inputQuicktimePath]);
      });

      it('returns input settings and input arguments for input with three added arguments', () => {
        const input = new FfmpegInput(inputFileMock);
        input.addOption('ss', 60);
        input.addOption('t', 10);
        input.addOption('c:v', 'prores');
        const testArguments = input.getCommandLine();
        expect(testArguments).to.deep.equal(['-ss', 60, '-t', 10, '-c:v', 'prores', '-i', inputQuicktimePath]);
      });
    });
  });
});
