/* eslint camelcase: "off" */
const FfmpegSettings = require('../FfmpegSettings/FfmpegSettings');
const { FfmpegTime } = require('../FfmpegValidators/FfmpegValidators');
const { FfmpegUnsupportedError } = require('../FfmpegErrors/FfmpegErrors');
require('dotenv').config();

module.exports = class FfmpegMediaFile {
  constructor() {
    this.ffmpegSettings = new FfmpegSettings();
  }

  /* Begin methods that map directly to Ffmpeg options */
  /**
   * Set the audio codec. This is an alias for -codec:a.
   * @param {string} codec
   */
  acodec(codec, supportedCodecs) {
    if (supportedCodecs.includes(codec)) {
      this.ffmpegSettings.addOption('c:a', codec);
      return this;
    }

    throw new FfmpegUnsupportedError(`Codec ${codec} is not supported on your version of ffmpeg`);
  }

  /**
   * As an input option, blocks all audio streams of a file from being filtered or being
   * automatically selected or mapped for any output. See -discard option to disable streams
   * individually.
   * As an output option, disables audio recording i.e. automatic selection or mapping of any audio
   * stream. For full manual control see the -map option.
   */
  an() {
    this.ffmpegSettings.addOption('an');
    return this;
  }

  /**
   * Enable bitexact mode for (de)muxer and (de/en)coder
   */
  bitexact() {
    this.ffmpegSettings.addOption('bitexact');
    return this;
  }

  /**
   * As an input option, blocks all data streams of a file from being filtered or being
   * automatically selected or mapped for any output. See -discard option to disable streams
   * individually.
   * As an output option, disables data recording i.e. automatic selection or mapping of any data
   * stream. For full manual control see the -map option.
   */
  dn() {
    this.ffmpegSettings.addOption('dn');
    return this;
  }

  /**
   * Force input or output file format. The format is normally auto detected for input files and
   * guessed from the file extension for output files, so this option is not needed in most cases.
   * @param {string} fmt
   */
  f(fmt) {
    this.ffmpegSettings.addOption('f', fmt);
    return this;
  }

  /**
    * When used as an input option (before -i), seeks in this input file to position. Note that in
    * most formats it is not possible to seek exactly, so ffmpeg will seek to the closest seek
    * point before position. When transcoding and -accurate_seek is enabled (the default), this
    * extra segment between the seek point and position will be decoded and discarded. When doing
    * stream copy or when -noaccurate_seek is used, it will be preserved.
    *
    * When used as an output option (before an output url), decodes but discards input until the
    * timestamps reach position.
    *
    * @param {number|string} position must be a time duration specification, see the Time duration
    * section in the ffmpeg-utils manual.
    * @throws {TypeError} if passed an invalid time specification
   */
  ss(position) {
    const time = new FfmpegTime(position);
    if (time.isValid()) {
      this.addOption('ss', time.get());
      return this;
    }
    throw new TypeError('ss requires an integer or string in the format [hh:]mm:ss[.MM] for an argument');
  }

  /**
    * When used as an input option, limit the duration of data read from the input file.
    * When used as an output option, stop writing the output after its duration reaches duration.
    * duration must be a time duration specification, see (ffmpeg-utils)the Time duration section in
    * the ffmpeg-utils manual.
    * @param {number|string} duration
    * @throws {TypeError} if passed an invalid time specification
   */
  t(duration) {
    const time = new FfmpegTime(duration);
    if (time.isValid()) {
      this.addOption('t', time.get());
      return this;
    }
    throw new TypeError('ss requires an integer or string in the format [hh:]mm:ss[.MM] for an argument');
  }

  /**
    * Stop writing the output or reading the input at position. to() and t() are mutually exclusive
    * and t() has priority.
    *
    * @param {number|string} duration must be a time durationspecification, see (ffmpeg-utils)the
    * Time duration section in the ffmpeg-utils manual.
    * @throws {TypeError} if passed an invalid time specification
  */
  to(duration) {
    const time = new FfmpegTime(duration);
    if (time.isValid()) {
      this.addOption('to', time.get());
      return this;
    }
    throw new TypeError('ss requires an integer or string in the format [hh:]mm:ss[.MM] for an argument');
  }

  /**
   * As an input option, blocks all subtitle streams of a file from being filtered or being
   * automatically selected or mapped for any output. See -discard option to disable streams
   * individually.
   * As an output option, disables subtitle recording i.e. automatic selection or mapping of
   * any subtitle stream. For full manual control see the -map option.
   */
  sn() {
    this.ffmpegSettings.addOption('sn');
    return this;
  }

  /**
   * Set SwScaler flags.
   * @param {string} flags
   */
  sws_flags(flags) {
    this.ffmpegSettings.addOption('sws_flags', flags);
    return this;
  }

  /** Set the video codec. This is an alias for -codec:v.
    *
    * @param {string} codec
   */
  vcodec(codec, supportedCodecs) {
    if (supportedCodecs.includes(codec)) {
      this.ffmpegSettings.addOption('c:v', codec);
      return this;
    }

    throw new FfmpegUnsupportedError(`Codec ${codec} is not supported on your version of ffmpeg`);
  }

  /** As an input option, blocks all video streams of a file from being filtered or being
   *  automatically selected or mapped for any output. See -discard option to disable streams
   *  individually.
   *  As an output option, disables video recording i.e. automatic selection or mapping of any video
   *  stream. For full manual control see the -map option.
  */
  vn() {
    this.ffmpegSettings.addOption('vn');
    return this;
  }
  /* End methods that map directly to Ffmpeg options */

  addOption(key, value) {
    this.ffmpegSettings.addOption(key, value);
    return this;
  }

  removeOption(option) {
    this.ffmpegSettings.removeOption(option);
    return this;
  }

  hasOption(option) {
    return this.ffmpegSettings.hasOption(option);
  }

  /**
   * Returns an array of command line arguments for ffmpeg, suitable for passing to
   * child_process.exec() or similar
   * @returns {string[]}
   */
  getCommandLine() {
    return this.ffmpegSettings.getCommandLine();
  }
};
