/* eslint-env mocha */
/* eslint-disable no-unused-expressions */
const { expect } = require('chai');
const { inputQuicktimePath, inputFileMock, inputFileMockWithManyStreams } = require('../../test/fixtures');
const FfmpegOutput = require('./FfmpegOutput');
const FfmpegInput = require('../FfmpegInput/FfmpegInput');
const { FfmpegUnsupportedError } = require('../FfmpegErrors/FfmpegErrors');
require('dotenv').config();

const { env: { DRAWTEXTFONT: DrawTextFont } } = process;

describe('FfmpegOutput', () => {
  describe('Unit Tests', () => {
    describe('Constructor', () => {
      const input = new FfmpegInput(inputFileMock);

      it('Instantiates empty FfmpegSettings', () => {
        const output = new FfmpegOutput(input);
        expect(output.ffmpegSettings.options).to.deep.equal({});
      });

      it('throws when called without an instance of FfmpegInput', () => {
        expect(() => new FfmpegOutput(inputFileMock)).to.throw();
      });
    });

    describe('getInputStreamFromMapString()', () => {
      const input = new FfmpegInput(inputFileMockWithManyStreams);

      it('finds all stream when called with "0"', () => {
        const output = new FfmpegOutput(input);
        const test = output.getInputStreamFromMapString('0');
        expect(test).to.deep.equal(input.streams);
      });

      it('finds first video stream with  "0:v:0"', () => {
        const output = new FfmpegOutput(input);
        const test = output.getInputStreamFromMapString('0:v:0');
        expect(test).to.deep.equal(input.streams[1]);
      });

      it('finds largest video stream with  "0:V"', () => {
        const output = new FfmpegOutput(input);
        const test = output.getInputStreamFromMapString('0:V');
        expect(test).to.deep.equal(input.streams[4]);
      });

      it('finds first audio stream with  "0:a:0"', () => {
        const output = new FfmpegOutput(input);
        const test = output.getInputStreamFromMapString('0:a:0');
        expect(test).to.deep.equal(input.streams[0]);
      });

      it('finds second audio stream with  "0:a:1"', () => {
        const output = new FfmpegOutput(input);
        const test = output.getInputStreamFromMapString('0:a:1');
        expect(test).to.deep.equal(input.streams[7]);
      });

      it('finds first data stream with  "0:d:0"', () => {
        const output = new FfmpegOutput(input);
        const test = output.getInputStreamFromMapString('0:d:0');
        expect(test).to.deep.equal(input.streams[2]);
      });

      it('finds first subtitle stream with "0:s:0"', () => {
        const output = new FfmpegOutput(input);
        const test = output.getInputStreamFromMapString('0:s:0');
        expect(test).to.deep.equal(input.streams[6]);
      });

      it('finds second data stream with  "0:d:1"', () => {
        const output = new FfmpegOutput(input);
        const test = output.getInputStreamFromMapString('0:d:1');
        expect(test).to.deep.equal(input.streams[3]);
      });

      it('finds the 9th stream with "0:8"', () => {
        const output = new FfmpegOutput(input);
        const test = output.getInputStreamFromMapString('0:8');
        expect(test).to.deep.equal(input.streams[8]);
      });

      it('throws with an invalid map string', () => {
        const output = new FfmpegOutput(input);
        expect(() => output.getInputStreamFromMapString('foo')).to.throw();
      });
    });

    describe('map()', () => {
      const input = new FfmpegInput(inputFileMock);
      it('adds FfmpegStream', () => {
        const inputStream = new FfmpegInput(inputFileMock).streams[0];
        const test = new FfmpegOutput(input);
        test.map(inputStream);
        expect(test.maps).to.deep.equal(['0:0']);
      });

      it('adds multiple arguments', () => {
        const test = new FfmpegOutput(input);
        test.map(input.streams[0], input.streams[1]);

        expect(test.maps).to.deep.equal(['0:0', '0:1']);
      });

      it('maps 0', () => {
        const test = new FfmpegOutput(input);
        test.map(0);
        expect(test.maps).to.deep.equal([0]);
      });

      it('throws when passed a number other than 0', () => {
        const test = new FfmpegOutput(input);
        expect(() => test.map(47)).to.throw();
      });

      it('throws with a number other than 0', () => {
        const test = new FfmpegOutput(input);
        expect(() => test.map(47)).to.throw();
      });
    });

    describe('getProgramArguments()', () => {
      const input = new FfmpegInput(inputFileMock);

      it('returns arguemnts for a program with stream, title, and number arguments', () => {
        const output = new FfmpegOutput(input);
        output.program([1, 0], { title: 'faketitle', program_num: 5 });
        const test = output.getProgramArguments();
        expect(test).to.deep.equal(['-program', 'title=faketitle:program_num=5:st=1:st=0']);
      });

      it('returns arguemnts for a program with only streams', () => {
        const output = new FfmpegOutput(input);
        output.program([1, 0]);
        const test = output.getProgramArguments();
        expect(test).to.deep.equal(['-program', 'st=1:st=0']);
      });
    });

    describe('getVideoFilterArguments()', () => {
      const input = new FfmpegInput(inputFileMock);
      it('returns a correct filter arguments array for a single filter without options', () => {
        const output = new FfmpegOutput(input);
        output.videoFilters = [{ name: 'decimate' }];
        const test = output.getVideoFilterArguments();
        expect(test).to.deep.equal(['-filter:v', 'decimate']);
      });

      it('returns a correct filter arguments array for a single filter with options', () => {
        const test = new FfmpegOutput(input);
        test.videoFilters = [{ name: 'scale', settings: { width: 1280, height: 720 } }];
        expect(test.getVideoFilterArguments()).to.deep.equal(['-filter:v', 'scale=width=1280:height=720']);
      });

      it('returns a correct filter arguments array for two filters, one with options', () => {
        const output = new FfmpegOutput(input);
        output.videoFilters = [{ name: 'decimate' }, {
          name: 'scale',
          settings: { width: 1280, height: 720 },
        }];
        const test = output.getVideoFilterArguments();
        expect(test).to.deep.equal(['-filter:v', 'decimate,scale=width=1280:height=720']);
      });

      it('returns an empty array with no filters set', () => {
        const output = new FfmpegOutput(input);
        const test = output.getVideoFilterArguments();
        expect(test).to.deep.equal([]);
      });
    });

    describe('getAudioFilterArguments()', () => {
      const input = new FfmpegInput(inputFileMock);

      it('returns a correct filter arguments array for a single filter with options', () => {
        const output = new FfmpegOutput(input);
        output.audioFilters = [{ name: 'asetpts' }];
        const test = output.getAudioFilterArguments();
        expect(test).to.deep.equal(['-filter:a', 'asetpts']);
      });

      it('returns a correct filter arguments array for two filters, one with options', () => {
        const output = new FfmpegOutput(input);
        output.audioFilters = [{
          name: 'aresample',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        }, {
          name: 'asetpts',
        }];

        const test = output.getAudioFilterArguments();
        expect(test).to.deep.equal(['-filter:a', 'aresample=fontsize=76:rate=59.94:tc24hmax=1:fontcolor=0xFFFFFFFF:boxcolor=0x000000FF:fix_bounds=true,asetpts']);
      });

      it('returns an empty array with no filters set', () => {
        const output = new FfmpegOutput(input);
        const test = output.getAudioFilterArguments();
        expect(test).to.deep.equal([]);
      });
    });

    describe('getComplexFilterArguments()', () => {
      const input = new FfmpegInput(inputFileMock);

      it('returns correct filter arguments with settings and input/output pads', () => {
        const output = new FfmpegOutput(input);
        output.complexFilters = [{
          name: 'amerge',
          inputPads: ['0:1', '0:2'],
          outputPads: ['aout'],
          settings: {
            inputs: '2',
          },
        }];
        const test = output.getComplexFilterArguments();
        expect(test).to.deep.equal(['-filter_complex', '[0:1][0:2]amerge=inputs=2[aout]']);
      });
    });

    describe('getMapArguments', () => {
      const input = new FfmpegInput(inputFileMock);
      it('generates correct map arguments', () => {
        const test = new FfmpegOutput(input);
        test.maps = ['0:1', '0:0'];
        expect(test.getMapArguments()).to.deep.equal(['-map', '0:1', '-map', '0:0']);
      });
    });
  });

  describe('Ffmpeg Options methods', () => {
    const input = new FfmpegInput(inputFileMock);

    describe('acodec()', () => {
      it('sets the audio codec for a supported codec', () => {
        const test = new FfmpegOutput(input);
        test.acodec('pcm_s24le');
        expect(test.ffmpegSettings.options).to.deep.equal({ 'c:a': 'pcm_s24le' });
      });

      it('throws when an unsupported codec is passed', () => {
        const test = new FfmpegOutput(input);
        expect(() => test.acodec('nonexistentaudiocodec')).to.throw();
      });
    });

    describe('af()', () => {
      it('parses a filtergraph string with no options', () => {
        const test = new FfmpegOutput(input);

        test.af('aresample');
        expect(test.audioFilters).to.deep.equal([{ name: 'aresample' }]);
      });

      it('parses a filtergraph string with filter options', () => {
        const test = new FfmpegOutput(input);

        test.af('aresample=fontsize=76:rate=59.94:tc24hmax=1:fontcolor=0xFFFFFFFF:boxcolor=0x000000FF:fix_bounds=true');
        expect(test.audioFilters).to.deep.equal([{
          name: 'aresample',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        }]);
      });

      it('adds a filtergraph object with one filter', () => {
        const test = new FfmpegOutput(input);

        test.af({
          name: 'aresample',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        });

        expect(test.audioFilters).to.deep.equal([{
          name: 'aresample',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        }]);
      });

      it('adds two filters, one with options, one without', () => {
        const test = new FfmpegOutput(input);

        test.af('aresample=fontsize=76:rate=59.94:tc24hmax=1:fontcolor=0xFFFFFFFF:boxcolor=0x000000FF:fix_bounds=true');
        test.af('asetpts');

        expect(test.audioFilters).to.deep.equal([
          {
            name: 'aresample',
            settings: {
              fontsize: '76',
              rate: '59.94',
              tc24hmax: '1',
              fontcolor: '0xFFFFFFFF',
              boxcolor: '0x000000FF',
              fix_bounds: 'true',
            },
          }, {
            name: 'asetpts',
          },
        ]);
      });

      it('adds a filtergraph object with two filters', () => {
        const test = new FfmpegOutput(input);

        test.af([{
          name: 'aresample',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        }, {
          name: 'asetpts',
          settings: {
            width: 720,
            height: 480,
          },
        }]);

        expect(test.audioFilters).to.deep.equal([
          {
            name: 'aresample',
            settings: {
              fontsize: '76',
              rate: '59.94',
              tc24hmax: '1',
              fontcolor: '0xFFFFFFFF',
              boxcolor: '0x000000FF',
              fix_bounds: 'true',
            },
          }, {
            name: 'asetpts',
            settings: {
              width: 720,
              height: 480,
            },
          },
        ]);
      });

      it('adds a second filtergraph', () => {
        const test = new FfmpegOutput(input);

        test.af({
          name: 'aresample',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        });

        test.af({
          name: 'asetpts',
          settings: {
            width: 720,
            height: 480,
          },
        });

        expect(test.audioFilters).to.deep.equal([
          {
            name: 'aresample',
            settings: {
              fontsize: '76',
              rate: '59.94',
              tc24hmax: '1',
              fontcolor: '0xFFFFFFFF',
              boxcolor: '0x000000FF',
              fix_bounds: 'true',
            },
          }, {
            name: 'asetpts',
            settings: {
              width: 720,
              height: 480,
            },
          },
        ]);
      });

      it('adds an array of filtergraph strings', () => {
        const test = new FfmpegOutput(input);

        test.af(['aresample=fontsize=76:rate=59.94:tc24hmax=1:fontcolor=0xFFFFFFFF:boxcolor=0x000000FF:fix_bounds=true', 'asetpts=width=720:height=480']);

        expect(test.audioFilters).to.deep.equal([
          {
            name: 'aresample',
            settings: {
              fontsize: '76',
              rate: '59.94',
              tc24hmax: '1',
              fontcolor: '0xFFFFFFFF',
              boxcolor: '0x000000FF',
              fix_bounds: 'true',
            },
          }, {
            name: 'asetpts',
            settings: {
              width: '720',
              height: '480',
            },
          },
        ]);
      });

      it('adds an array of filtergraph objects', () => {
        const test = new FfmpegOutput(input);

        test.af([{
          name: 'aresample',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        }, {
          name: 'asetpts',
          settings: {
            width: 720,
            height: 480,
          },
        }]);

        expect(test.audioFilters).to.deep.equal([
          {
            name: 'aresample',
            settings: {
              fontsize: '76',
              rate: '59.94',
              tc24hmax: '1',
              fontcolor: '0xFFFFFFFF',
              boxcolor: '0x000000FF',
              fix_bounds: 'true',
            },
          }, {
            name: 'asetpts',
            settings: {
              width: 720,
              height: 480,
            },
          },
        ]);
      });

      it('throws TypeError with a number argument', () => {
        const test = new FfmpegOutput(input);
        expect(() => test.af(5)).to.throw(TypeError);
      });

      it('throws FfmpegUnsupportedError with an unsupported filter', () => {
        const test = new FfmpegOutput(input);
        expect(() => test.af({ fields: { settings: { order: 'bottom' } } })).to.throw(FfmpegUnsupportedError);
      });
    });

    describe('aframes', () => {
      it('accepts an integer', () => {
        const test = new FfmpegOutput(input);

        test.aframes(45);
        expect(test.hasOption('frames:a')).to.be.true;
      });

      it('accepts an integer stored as string', () => {
        const test = new FfmpegOutput(input);

        test.aframes('75');
        expect(test.hasOption('frames:a')).to.be.true;
      });

      it('throws with invalid argument', () => {
        const test = new FfmpegOutput(input);

        expect(() => test.aframes('fifty').to.throw());
      });
    });

    describe('aq()', () => {
      it('accepts an integer', () => {
        const test = new FfmpegOutput(input);

        test.aq(45);
        expect(test.hasOption('q:a')).to.be.true;
      });

      it('does nothing with no argument', () => {
        const test = new FfmpegOutput(input);

        test.aq();
        expect(test.hasOption('frames:a')).to.be.false;
      });
    });

    describe('attach()', () => {
      it('attaches a file', () => {
        const output = new FfmpegOutput(input);
        output.attach(inputQuicktimePath);
        expect(output.hasOption('attach')).to.be.true;
      });

      it('throws when passed a nonexistent file', () => {
        const output = new FfmpegOutput(input);
        expect(() => output.attach('foo')).to.throw();
      });
    });

    describe('dframes()', () => {
      it('adds dframes to options', () => {
        const output = new FfmpegOutput(input);
        output.dframes(56);
        expect(output.hasOption('frames:d')).to.be.true;
      });

      it('throws when passed a string', () => {
        const output = new FfmpegOutput(input);
        expect(() => output.dframes('foo')).to.throw();
      });
    });

    describe('filter_complex()', () => {
      it('parses a filtergrapb string with no options', () => {
        const test = new FfmpegOutput(input);

        test.filter_complex('aresample');
        expect(test.complexFilters).to.deep.equal([{ name: 'aresample' }]);
      });

      it('parses a filtergraph string with filter options', () => {
        const test = new FfmpegOutput(input);

        test.filter_complex('aresample=fontsize=76:rate=59.94:tc24hmax=1:fontcolor=0xFFFFFFFF:boxcolor=0x000000FF:fix_bounds=true');
        expect(test.complexFilters).to.deep.equal([{
          name: 'aresample',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        }]);
      });

      it('adds a filtergraph string with input and output pads', () => {
        const output = new FfmpegOutput(input);
        output.filter_complex('[0:1][0:7]amerge=inputs=2[aout]');
        const test = output.complexFilters;

        expect(test).to.deep.equal([{
          name: 'amerge',
          inputPads: ['0:1', '0:7'],
          outputPads: ['aout'],
          settings: {
            inputs: '2',
          },
        }]);
      });

      it('adds a filtergraph object with one filter', () => {
        const test = new FfmpegOutput(input);

        test.filter_complex({
          name: 'aresample',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        });

        expect(test.complexFilters).to.deep.equal([{
          name: 'aresample',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        }]);
      });

      it('adds two filters, one with options, one without', () => {
        const test = new FfmpegOutput(input);

        test.filter_complex('aresample=fontsize=76:rate=59.94:tc24hmax=1:fontcolor=0xFFFFFFFF:boxcolor=0x000000FF:fix_bounds=true');
        test.filter_complex('asetpts');

        expect(test.complexFilters).to.deep.equal([
          {
            name: 'aresample',
            settings: {
              fontsize: '76',
              rate: '59.94',
              tc24hmax: '1',
              fontcolor: '0xFFFFFFFF',
              boxcolor: '0x000000FF',
              fix_bounds: 'true',
            },
          }, {
            name: 'asetpts',
          },
        ]);
      });

      it('adds a filtergraph object with two filters', () => {
        const test = new FfmpegOutput(input);

        test.filter_complex([{
          name: 'aresample',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        }, {
          name: 'asetpts',
          settings: {
            width: 720,
            height: 480,
          },
        }]);

        expect(test.complexFilters).to.deep.equal([
          {
            name: 'aresample',
            settings: {
              fontsize: '76',
              rate: '59.94',
              tc24hmax: '1',
              fontcolor: '0xFFFFFFFF',
              boxcolor: '0x000000FF',
              fix_bounds: 'true',
            },
          }, {
            name: 'asetpts',
            settings: {
              width: 720,
              height: 480,
            },
          },
        ]);
      });

      it('adds a second filtergraph', () => {
        const test = new FfmpegOutput(input);

        test.filter_complex({
          name: 'aresample',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        });

        test.filter_complex({
          name: 'asetpts',
          settings: {
            width: 720,
            height: 480,
          },
        });

        expect(test.complexFilters).to.deep.equal([
          {
            name: 'aresample',
            settings: {
              fontsize: '76',
              rate: '59.94',
              tc24hmax: '1',
              fontcolor: '0xFFFFFFFF',
              boxcolor: '0x000000FF',
              fix_bounds: 'true',
            },
          }, {
            name: 'asetpts',
            settings: {
              width: 720,
              height: 480,
            },
          },
        ]);
      });

      it('throws TypeError with a number argument', () => {
        const test = new FfmpegOutput(input);
        expect(() => test.filter_complex(5)).to.throw(TypeError);
      });

      it('throws FfmpegUnsupportedError with an unsupported filter', () => {
        const test = new FfmpegOutput(input);
        expect(() => test.filter_complex({ fields: { order: 'bottom' } })).to.throw(FfmpegUnsupportedError);
      });
    });

    describe('frames()', () => {
      it('adds frames to options', () => {
        const output = new FfmpegOutput(input);
        output.frames(75);
        expect(output.hasOption('frames')).to.be.true;
      });

      it('throws when passed a string', () => {
        const output = new FfmpegOutput(input);
        expect(() => output.dframes('foo')).to.throw();
      });
    });

    describe('fs()', () => {
      it('adds fs to options', () => {
        const output = new FfmpegOutput(input);
        output.fs(1024);
        expect(output.hasOption('fs')).to.be.true;
      });

      it('throws when passed a string', () => {
        const output = new FfmpegOutput(input);
        expect(() => output.fs('foo')).to.throw();
      });
    });

    describe('program()', () => {
      it('adds a program with only a string argument', () => {
        const test = new FfmpegOutput(input);
        test.program(0);
        expect(test.programs).to.deep.equal(['st=0']);
      });

      it('adds a program with a string, title, and program_num argument', () => {
        const test = new FfmpegOutput(input);
        test.program('0', { title: 'faketitle', program_num: 1 });
        expect(test.programs).to.deep.equal(['title=faketitle:program_num=1:st=0']);
      });

      it('adds an array of streams', () => {
        const test = new FfmpegOutput(input);
        test.program(['0', 1]);
        expect(test.programs).to.deep.equal(['st=0:st=1']);
      });
    });

    describe('q()', () => {
      it('adds qscale to options', () => {
        const output = new FfmpegOutput(input);
        output.q(1024);
        expect(output.hasOption('qscale')).to.be.true;
      });

      it('throws when passed a string', () => {
        const output = new FfmpegOutput(input);
        expect(() => output.q('foo')).to.throw();
      });
    });

    describe('qscale()', () => {
      it('adds qscale to options', () => {
        const output = new FfmpegOutput(input);
        output.qscale(1024);
        expect(output.hasOption('qscale')).to.be.true;
      });

      it('throws when passed a string', () => {
        const output = new FfmpegOutput(input);
        expect(() => output.qscale('foo')).to.throw();
      });
    });

    describe('sample_fmt()', () => {
      it('adds a valid sample_fmt', () => {
        const output = new FfmpegOutput(input);
        output.sample_fmt('s16');
        expect(output.hasOption('sample_fmt')).to.be.true;
      });

      it('does nothing when called with no argument', () => {
        const output = new FfmpegOutput(input);
        output.sample_fmt();
        expect(output.hasOption('sample_fmt')).to.be.false;
      });

      it('throws when given an invalid sample_fmt', () => {
        const output = new FfmpegOutput(input);
        expect(() => output.sample_fmt('foo')).to.throw(FfmpegUnsupportedError);
      });
    });

    describe('target()', () => {
      it('adds a target of "dv"', () => {
        const output = new FfmpegOutput(input);
        output.target('dv');
        expect(output.hasOption('target')).to.be.true;
      });

      it('adds a target of film-dvd', () => {
        const output = new FfmpegOutput(input);
        output.target('film-dvd');
        expect(output.hasOption('target')).to.be.true;
      });

      it('doesn\'t add an invalid target', () => {
        const output = new FfmpegOutput(input);
        expect(() => output.target('ntsc-bluray')).to.throw(TypeError);
      });
    });

    describe('vcodec()', () => {
      it('sets the video codec for a supported codec', () => {
        const test = new FfmpegOutput(input);
        test.vcodec('rawvideo');
        expect(test.hasOption('c:v')).to.be.true;
      });

      it('throws when an unsupported codec is passed', () => {
        const test = new FfmpegOutput(input);
        expect(() => test.vcodec('nonexistentvideocodec')).to.throw();
      });
    });

    describe('vf()', () => {
      it('parses a filtergraph string with no options', () => {
        const test = new FfmpegOutput(input);

        test.vf('decimate');
        expect(test.videoFilters).to.deep.equal([{ name: 'decimate' }]);
      });

      it('parses a filtergraph string with filter options', () => {
        const test = new FfmpegOutput(input);

        test.vf('drawtext=fontsize=76:rate=59.94:tc24hmax=1:fontcolor=0xFFFFFFFF:boxcolor=0x000000FF:fix_bounds=true');
        expect(test.videoFilters).to.deep.equal([{
          name: 'drawtext',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        }]);
      });

      it('adds a filtergraph object with one filter', () => {
        const test = new FfmpegOutput(input);

        test.vf({
          name: 'drawtext',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        });

        expect(test.videoFilters).to.deep.equal([{
          name: 'drawtext',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        }]);
      });

      it('adds two filters, one with options, one without', () => {
        const test = new FfmpegOutput(input);

        test.vf('drawtext=fontsize=76:rate=59.94:tc24hmax=1:fontcolor=0xFFFFFFFF:boxcolor=0x000000FF:fix_bounds=true');
        test.vf('decimate');

        expect(test.videoFilters).to.deep.equal([
          {
            name: 'drawtext',
            settings: {
              fontsize: '76',
              rate: '59.94',
              tc24hmax: '1',
              fontcolor: '0xFFFFFFFF',
              boxcolor: '0x000000FF',
              fix_bounds: 'true',
            },
          }, {
            name: 'decimate',
          },
        ]);
      });

      it('adds a filtergraph object array with two filters', () => {
        const test = new FfmpegOutput(input);

        test.vf([{
          name: 'drawtext',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        }, {
          name: 'scale',
          settings: {
            width: 720,
            height: 480,
          },
        }]);

        expect(test.videoFilters).to.deep.equal([
          {
            name: 'drawtext',
            settings: {
              fontsize: '76',
              rate: '59.94',
              tc24hmax: '1',
              fontcolor: '0xFFFFFFFF',
              boxcolor: '0x000000FF',
              fix_bounds: 'true',
            },
          }, {
            name: 'scale',
            settings: {
              width: 720,
              height: 480,
            },
          },
        ]);
      });

      it('adds a second filtergraph', () => {
        const test = new FfmpegOutput(input);

        test.vf({
          name: 'drawtext',
          settings: {
            fontsize: '76',
            rate: '59.94',
            tc24hmax: '1',
            fontcolor: '0xFFFFFFFF',
            boxcolor: '0x000000FF',
            fix_bounds: 'true',
          },
        });

        test.vf({
          name: 'scale',
          settings: {
            width: 720,
            height: 480,
          },
        });

        expect(test.videoFilters).to.deep.equal([
          {
            name: 'drawtext',
            settings: {
              fontsize: '76',
              rate: '59.94',
              tc24hmax: '1',
              fontcolor: '0xFFFFFFFF',
              boxcolor: '0x000000FF',
              fix_bounds: 'true',
            },
          }, {
            name: 'scale',
            settings: {
              width: 720,
              height: 480,
            },
          },
        ]);
      });

      it('throws TypeError with a number argument', () => {
        const test = new FfmpegOutput(input);
        expect(() => test.vf(5)).to.throw(TypeError);
      });

      it('throws FfmpegUnsupportedError with an unsupported filter', () => {
        const test = new FfmpegOutput(input);
        expect(() => test.vf({ fields: { order: 'bottom' } })).to.throw(FfmpegUnsupportedError);
      });
    });

    describe('vframes()', () => {
      it('sets frames:v with an integer argument', () => {
        const test = new FfmpegOutput(input);
        test.vframes(4);
        expect(test.hasOption('frames:v')).to.be.true;
      });

      it('sets frames:v with an integer stored as string argument', () => {
        const test = new FfmpegOutput(input);
        test.vframes('4');
        expect(test.hasOption('frames:v')).to.be.true;
      });

      it('throws when a non-number is passed', () => {
        const test = new FfmpegOutput(input);
        expect(() => test.vframes('nonsense')).to.throw();
      });
    });

    describe('vtag()', () => {
      it('sets the video codec for a supported codec', () => {
        const test = new FfmpegOutput(input);
        test.vtag('tag');
        expect(test.hasOption('tag:v')).to.be.true;
      });

      it('does nothing with no argument', () => {
        const test = new FfmpegOutput(input);
        test.vtag();
        expect(test.hasOption('tag:v')).to.be.false;
      });
    });
  });

  describe('Inheritance/Composition tests', () => {
    const input = new FfmpegInput(inputFileMock);
    describe('FfmpegSettings', () => {
      describe('addOption()', () => {
        it('adds an argument', () => {
          const test = new FfmpegOutput(input);
          test.addOption('ss', 60);
          expect(test.hasOption('ss')).to.be.true;
        });
      });

      describe('removeOption()', () => {
        it('removes an argument', () => {
          const test = new FfmpegOutput(input);
          test.addOption('ss', 60);
          expect(test.hasOption('ss'), 'Failed to set initial values to be deleted').to.be.true;
          test.removeOption('ss');
          expect(test.hasOption('ss')).to.be.false;
        });
      });
    });
  });

  describe('Integration Tests', () => {
    const input = new FfmpegInput(inputFileMock);

    describe('resize()', () => {
      it('accepts a width and height', () => {
        const test = new FfmpegOutput(input);
        test.resize({ width: 1920, height: 1080 });
        expect(test.videoFilters).to.deep.equal([{
          name: 'scale',
          settings: {
            width: 1920,
            height: 1080,
            sws_flags: 'fast_bilinear',
          },
        }]);
      });

      it('accepts a width and height', () => {
        const test = new FfmpegOutput(input);
        test.resize({ width: 1920 });
        expect(test.videoFilters).to.deep.equal([{
          name: 'scale',
          settings: {
            width: 1920,
            height: 1080,
            sws_flags: 'fast_bilinear',
          },
        }]);
      });

      it('accepts a width and height', () => {
        const test = new FfmpegOutput(input);
        test.resize({ height: 1080 });
        expect(test.videoFilters).to.deep.equal([{
          name: 'scale',
          settings: {
            width: 1920,
            height: 1080,
            sws_flags: 'fast_bilinear',
          },
        }]);
      });
    });

    describe('addTimecodeBurn()', () => {
      it('with adds the filter when passed ("01:00:00:00", r: 30)', () => {
        const test = new FfmpegOutput(input);
        test.addTimecodeBurn('01:00:00:00', 30);
        expect(test.videoFilters).to.deep.equal([{
          name: 'drawtext',
          settings: {
            timecode: '01\\\\:00\\\\:00\\\\:00',
            r: 30,
            fontfile: DrawTextFont,
            fontsize: '(h/16)',
            fontcolor: 'white',
            box: true,
            boxcolor: 'black',
            x: '((w-tw)-(w/10))',
            y: '(h/10)',
          },
        }]);
      });

      it('with adds the filter when passed ("01:00:00;00", "30000/1001")', () => {
        const test = new FfmpegOutput(input);
        test.addTimecodeBurn('01:00:00;00', '30000/1001');
        expect(test.videoFilters).to.deep.equal([{
          name: 'drawtext',
          settings: {
            timecode: '01\\\\:00\\\\:00\\;00',
            r: '30000/1001',
            fontfile: DrawTextFont,
            fontsize: '(h/16)',
            fontcolor: 'white',
            box: true,
            boxcolor: 'black',
            x: '((w-tw)-(w/10))',
            y: '(h/10)',
          },
        }]);
      });

      it('sets all when only given ("01:00:00:00", pal)', () => {
        const test = new FfmpegOutput(input);
        test.addTimecodeBurn('01:00:00;00', 'pal');
        expect(test.videoFilters).to.deep.equal([{
          name: 'drawtext',
          settings: {
            timecode: '01\\\\:00\\\\:00\\;00',
            r: 'pal',
            fontfile: DrawTextFont,
            fontsize: '(h/16)',
            fontcolor: 'white',
            box: true,
            boxcolor: 'black',
            x: '((w-tw)-(w/10))',
            y: '(h/10)',
          },
        }]);
      });

      it('correctly formats { timecode: "01:00" }', () => {
        const test = new FfmpegOutput(input);

        test.addTimecodeBurn('01:00', 30);
        expect(test.videoFilters).to.deep.equal([{
          name: 'drawtext',
          settings: {
            timecode: '01\\\\:00\\\\:00\\\\:00',
            r: 30,
            fontfile: DrawTextFont,
            fontsize: '(h/16)',
            fontcolor: 'white',
            box: true,
            boxcolor: 'black',
            x: '((w-tw)-(w/10))',
            y: '(h/10)',
          },
        }]);
      });

      it('throws with 01000000', () => {
        const test = new FfmpegOutput(input);
        expect(() => test.addTimecodeBurn('01000000', 30)).to.throw;
      });
    });

    describe('getCommandLine()', () => {
      it('gets full arguments array', () => {
        const test = new FfmpegOutput(input)
          .map(['0:1', '0:0'])
          .vf({ name: 'scale', settings: { width: 1280, height: 720 } })
          .vf({ name: 'field', settings: { type: 'bottom' } })
          .vcodec('libx264')
          .acodec('aac');
        expect(test.getCommandLine()).to.deep.equal(['-c:v', 'libx264', '-c:a', 'aac', '-filter:v', 'scale=width=1280:height=720,field=type=bottom', '-map', '0:1', '-map', '0:0']);
      });
    });
  });
});
