const childProcess = require('child_process');
require('dotenv').config();

const { env: { FFMPEG_BIN } } = process;
const [FfmpegBin, ...FfmpegDefaultArguments] = FFMPEG_BIN.split(' ');

const codecParsingPattern = /^\s+(V|A).{5}\s((?:\w|-)+)\s+(.*)$/gm;
const filterParsingPattern = /^\s.{4}(\w+)\s+(?:(\w+|\|)+)->(?:(\w+|\|)+)\s+(.*)/gm;
const buildConfigurationParsingPattern = /--([\w-]+)(=)?([\w-/.]+)?/g;
const sampleFmtsPattern = /^(\w+)\s+\d+/gm;

module.exports = {
  parseFfmpegVersionNumber: function parseFfmpegVersionNumber() {
    const stdout = childProcess.execFileSync(FfmpegBin, FfmpegDefaultArguments.concat(['-version', '-hide_banner']));
    const [, match] = /ffmpeg version (\S+).*/.exec(stdout);
    return parseFloat(match, 10);
  },

  parseFfmpegBuildOutput: function parseFfmpegBuildOutput() {
    const buildOutput = childProcess.execFileSync(FfmpegBin, FfmpegDefaultArguments.concat(['-version', '-hide_banner']));
    const data = {};
    let matches = buildConfigurationParsingPattern.exec(buildOutput);

    while (matches) {
      const [, option, equalSign, setting] = matches;

      if (equalSign) {
        data[option] = setting || ''; // eslint-disable-line no-param-reassign
      } else {
        data[option] = true; // eslint-disable-line no-param-reassign
      }
      matches = buildConfigurationParsingPattern.exec(buildOutput);
    }

    Object.freeze(data);
    return data;
  },

  parseFfmpegEncoderOutput: function parseFfmpegEncoderOutput() {
    const encodersOutput = childProcess.execFileSync(FfmpegBin, FfmpegDefaultArguments.concat(['-encoders', '-hide_banner']));
    const ffmpegVideoEncoders = { copy: 'Stream Copy' };
    const ffmpegAudioEncoders = { copy: 'Stream Copy' };
    let matches = codecParsingPattern.exec(encodersOutput);

    while (matches) {
      const [, type, code, description] = matches;

      if (type === 'V') {
        ffmpegVideoEncoders[code] = description; // eslint-disable-line no-param-reassign
      } else if (type === 'A') {
        ffmpegAudioEncoders[code] = description; // eslint-disable-line no-param-reassign
      }
      matches = codecParsingPattern.exec(encodersOutput);
    }

    Object.freeze(ffmpegVideoEncoders);
    Object.freeze(ffmpegAudioEncoders);
    return { ffmpegVideoEncoders, ffmpegAudioEncoders };
  },

  parseFfmpegDecoderOutput: function parseFfmpegEncoderOutput() {
    const decodersOutput = childProcess.execFileSync(FfmpegBin, FfmpegDefaultArguments.concat(['-decoders', '-hide_banner']));
    const ffmpegVideoDecoders = {};
    const ffmpegAudioDecoders = {};
    let matches = codecParsingPattern.exec(decodersOutput);

    while (matches) {
      const [, type, code, description] = matches;

      if (type === 'V') {
        ffmpegVideoDecoders[code] = description; // eslint-disable-line no-param-reassign
      } else if (type === 'A') {
        ffmpegAudioDecoders[code] = description; // eslint-disable-line no-param-reassign
      }
      matches = codecParsingPattern.exec(decodersOutput);
    }

    Object.freeze(ffmpegVideoDecoders);
    Object.freeze(ffmpegAudioDecoders);
    return { ffmpegVideoDecoders, ffmpegAudioDecoders };
  },

  parseFfmpegSampleFmtsOutput: function parseFfmpegSampleFmtsOutput() {
    const sampleFmtOutput = childProcess.execFileSync(FfmpegBin, FfmpegDefaultArguments.concat(['-sample_fmts', '-hide_banner']));
    let matches = sampleFmtsPattern.exec(sampleFmtOutput);

    const sampleFmts = [];
    while (matches) {
      const [, sampleFmt] = matches;
      sampleFmts.push(sampleFmt);
      matches = sampleFmtsPattern.exec(sampleFmtOutput);
    }

    Object.freeze(sampleFmts);
    return sampleFmts;
  },

  decodeFilterType: function decodeFilterType(code) {
    if (/V+/.test(code)) return 'video';
    if (/A+/.test(code)) return 'audio';
    if (code === 'N') return 'dynamic';

    return 'null';
  },

  parseFfmpegFilterOutput: function parseFfmpegFilterOutput() {
    const ffmpegFilters = {};
    const filterOutput = childProcess.execFileSync(FfmpegBin, FfmpegDefaultArguments.concat(['-filters', '-hide_banner']));

    let matches = filterParsingPattern.exec(filterOutput);

    while (matches) {
      const [, code, input, output, description] = matches;
      ffmpegFilters[code] = {
        input: this.decodeFilterType(input),
        output: this.decodeFilterType(output),
        description,
      };
      matches = filterParsingPattern.exec(filterOutput);
    }
    Object.freeze(ffmpegFilters);

    return ffmpegFilters;
  },

  FfmpegResizePresets: {
    ntsc: { width: 720, height: 480 },
    pal: { width: 720, height: 576 },
    qntsc: { width: 352, height: 240 },
    qpal: { width: 352, height: 288 },
    sntsc: { width: 640, height: 480 },
    spal: { width: 768, height: 576 },
    film: { width: 352, height: 240 },
    'ntsc-film': { width: 352, height: 240 },
    sqcif: { width: 128, height: 96 },
    qcif: { width: 176, height: 144 },
    cif: { width: 352, height: 288 },
    '4cif': { width: 704, height: 576 },
    '16cif': { width: 1408, height: 1152 },
    qqvga: { width: 160, height: 120 },
    qvga: { width: 320, height: 240 },
    vga: { width: 640, height: 480 },
    svga: { width: 800, height: 600 },
    xga: { width: 1024, height: 768 },
    uxga: { width: 1600, height: 1200 },
    qxga: { width: 2048, height: 1536 },
    sxga: { width: 1280, height: 1024 },
    qsxga: { width: 2560, height: 2048 },
    hsxga: { width: 5120, height: 4096 },
    wvga: { width: 852, height: 480 },
    wxga: { width: 1366, height: 768 },
    wsxga: { width: 1600, height: 1024 },
    wuxga: { width: 1920, height: 1200 },
    woxga: { width: 2560, height: 1600 },
    wqsxga: { width: 3200, height: 2048 },
    wquxga: { width: 3840, height: 2400 },
    whsxga: { width: 6400, height: 4096 },
    whuxga: { width: 7680, height: 4800 },
    cga: { width: 320, height: 200 },
    ega: { width: 640, height: 350 },
    hd480: { width: 852, height: 480 },
    hd720: { width: 1280, height: 720 },
    hd1080: { width: 1920, height: 1080 },
    '2k': { width: 2048, height: 1080 },
    '2kflat': { width: 1998, height: 1080 },
    '2kscope': { width: 2048, height: 858 },
    '4k': { width: 4096, height: 2160 },
    '4kflat': { width: 3996, height: 2160 },
    '4kscope': { width: 4096, height: 1716 },
    nhd: { width: 640, height: 360 },
    hqvga: { width: 240, height: 160 },
    wqvga: { width: 400, height: 240 },
    fwqvga: { width: 432, height: 240 },
    hvga: { width: 480, height: 320 },
    qhd: { width: 960, height: 540 },
    '2kdci': { width: 2048, height: 1080 },
    '4kdci': { width: 4096, height: 2160 },
    uhd2160: { width: 3840, height: 2160 },
    uhd4320: { width: 7680, height: 4320 },
  },

  FfmpegFrameratePresets: {
    ntsc: '30000/1001',
    pal: '25/1',
    qntsc: '30000/1001',
    qpal: '25/1',
    sntsc: '30000/1001',
    spal: '25/1',
    film: '24/1',
    'ntsc-film': '24000/1001',
  },

  FfmpegColorPresets: {
    AliceBlue: '0xF0F8FF',
    AntiqueWhite: '0xFAEBD7',
    Aqua: '0x00FFFF',
    Aquamarine: '0x7FFFD4',
    Azure: '0xF0FFFF',
    Beige: '0xF5F5DC',
    Bisque: '0xFFE4C4',
    Black: '0x000000',
    BlanchedAlmond: '0xFFEBCD',
    Blue: '0x0000FF',
    BlueViolet: '0x8A2BE2',
    Brown: '0xA52A2A',
    BurlyWood: '0xDEB887',
    CadetBlue: '0x5F9EA0',
    Chartreuse: '0x7FFF00',
    Chocolate: '0xD2691E',
    Coral: '0xFF7F50',
    CornflowerBlue: '0x6495ED',
    Cornsilk: '0xFFF8DC',
    Crimson: '0xDC143C',
    Cyan: '0x00FFFF',
    DarkBlue: '0x00008B',
    DarkCyan: '0x008B8B',
    DarkGoldenRod: '0xB8860B',
    DarkGray: '0xA9A9A9',
    DarkGreen: '0x006400',
    DarkKhaki: '0xBDB76B',
    DarkMagenta: '0x8B008B',
    DarkOliveGreen: '0x556B2F',
    Darkorange: '0xFF8C00',
    DarkOrchid: '0x9932CC',
    DarkRed: '0x8B0000',
    DarkSalmon: '0xE9967A',
    DarkSeaGreen: '0x8FBC8F',
    DarkSlateBlue: '0x483D8B',
    DarkSlateGray: '0x2F4F4F',
    DarkTurquoise: '0x00CED1',
    DarkViolet: '0x9400D3',
    DeepPink: '0xFF1493',
    DeepSkyBlue: '0x00BFFF',
    DimGray: '0x696969',
    DodgerBlue: '0x1E90FF',
    FireBrick: '0xB22222',
    FloralWhite: '0xFFFAF0',
    ForestGreen: '0x228B22',
    Fuchsia: '0xFF00FF',
    Gainsboro: '0xDCDCDC',
    GhostWhite: '0xF8F8FF',
    Gold: '0xFFD700',
    GoldenRod: '0xDAA520',
    Gray: '0x808080',
    Green: '0x008000',
    GreenYellow: '0xADFF2F',
    HoneyDew: '0xF0FFF0',
    HotPink: '0xFF69B4',
    IndianRed: '0xCD5C5C',
    Indigo: '0x4B0082',
    Ivory: '0xFFFFF0',
    Khaki: '0xF0E68C',
    Lavender: '0xE6E6FA',
    LavenderBlush: '0xFFF0F5',
    LawnGreen: '0x7CFC00',
    LemonChiffon: '0xFFFACD',
    LightBlue: '0xADD8E6',
    LightCoral: '0xF08080',
    LightCyan: '0xE0FFFF',
    LightGoldenRodYellow: '0xFAFAD2',
    LightGreen: '0x90EE90',
    LightGrey: '0xD3D3D3',
    LightPink: '0xFFB6C1',
    LightSalmon: '0xFFA07A',
    LightSeaGreen: '0x20B2AA',
    LightSkyBlue: '0x87CEFA',
    LightSlateGray: '0x778899',
    LightSteelBlue: '0xB0C4DE',
    LightYellow: '0xFFFFE0',
    Lime: '0x00FF00',
    LimeGreen: '0x32CD32',
    Linen: '0xFAF0E6',
    Magenta: '0xFF00FF',
    Maroon: '0x800000',
    MediumAquaMarine: '0x66CDAA',
    MediumBlue: '0x0000CD',
    MediumOrchid: '0xBA55D3',
    MediumPurple: '0x9370D8',
    MediumSeaGreen: '0x3CB371',
    MediumSlateBlue: '0x7B68EE',
    MediumSpringGreen: '0x00FA9A',
    MediumTurquoise: '0x48D1CC',
    MediumVioletRed: '0xC71585',
    MidnightBlue: '0x191970',
    MintCream: '0xF5FFFA',
    MistyRose: '0xFFE4E1',
    Moccasin: '0xFFE4B5',
    NavajoWhite: '0xFFDEAD',
    Navy: '0x000080',
    OldLace: '0xFDF5E6',
    Olive: '0x808000',
    OliveDrab: '0x6B8E23',
    Orange: '0xFFA500',
    OrangeRed: '0xFF4500',
    Orchid: '0xDA70D6',
    PaleGoldenRod: '0xEEE8AA',
    PaleGreen: '0x98FB98',
    PaleTurquoise: '0xAFEEEE',
    PaleVioletRed: '0xD87093',
    PapayaWhip: '0xFFEFD5',
    PeachPuff: '0xFFDAB9',
    Peru: '0xCD853F',
    Pink: '0xFFC0CB',
    Plum: '0xDDA0DD',
    PowderBlue: '0xB0E0E6',
    Purple: '0x800080',
    Red: '0xFF0000',
    RosyBrown: '0xBC8F8F',
    RoyalBlue: '0x4169E1',
    SaddleBrown: '0x8B4513',
    Salmon: '0xFA8072',
    SandyBrown: '0xF4A460',
    SeaGreen: '0x2E8B57',
    SeaShell: '0xFFF5EE',
    Sienna: '0xA0522D',
    Silver: '0xC0C0C0',
    SkyBlue: '0x87CEEB',
    SlateBlue: '0x6A5ACD',
    SlateGray: '0x708090',
    Snow: '0xFFFAFA',
    SpringGreen: '0x00FF7F',
    SteelBlue: '0x4682B4',
    Tan: '0xD2B48C',
    Teal: '0x008080',
    Thistle: '0xD8BFD8',
    Tomato: '0xFF6347',
    Turquoise: '0x40E0D0',
    Violet: '0xEE82EE',
    Wheat: '0xF5DEB3',
    White: '0xFFFFFF',
    WhiteSmoke: '0xF5F5F5',
    Yellow: '0xFFFF00',
    YellowGreen: '0x9ACD32',
  },
};
