/* eslint-env mocha */
/* eslint-disable no-unused-expressions */
const { expect } = require('chai');
const FfmpegSettings = require('./FfmpegSettings');
const { inputQuicktimePath } = require('../../test/fixtures');
require('dotenv').config();

describe('FfmpegSettings', () => {
  describe('Unit Tests', () => {
    describe('addOption()', () => {
      it('adds an argument without a hyphen', () => {
        const test = new FfmpegSettings();
        test.addOption('ss', 60);
        expect(test.options).to.deep.equal({ ss: 60 });
      });

      it('adds and corrects an option with no setting with a hyphen', () => {
        const test = new FfmpegSettings();
        test.addOption('-vn');
        expect(test.options).to.deep.equal({ vn: null });
      });

      it('removes first argument of two', () => {
        const test = new FfmpegSettings();
        test.addOption('ss', 60);
        test.addOption('t', 50);
        expect(test.options).to.deep.equal({ ss: 60, t: 50 });
        test.removeOption('ss');
        expect(test.options).to.deep.equal({ t: 50 });
      });

      it('removes second argument of two', () => {
        const test = new FfmpegSettings();
        test.addOption('ss', 60);
        test.addOption('t', 50);
        expect(test.options).to.deep.equal({ ss: 60, t: 50 }, 'Failed to set initial values to be deleted');
        test.removeOption('t');
        expect(test.options).to.deep.equal({ ss: 60 });
      });

      it('removes first argument of three', () => {
        const test = new FfmpegSettings();
        test.addOption('ss', 60);
        test.addOption('t', 50);
        test.addOption('c:v', 'h264');
        expect(test.options).to.deep.equal({ ss: 60, t: 50, 'c:v': 'h264' });
        test.removeOption('ss');
        expect(test.hasOption('ss')).to.be.false;
      });

      it('removes second argument of three', () => {
        const test = new FfmpegSettings();
        test.addOption('ss', 60);
        test.addOption('t', 50);
        test.addOption('c:v', 'h264');
        expect(test.options).to.deep.equal({ ss: 60, t: 50, 'c:v': 'h264' });
        test.removeOption('t');
        expect(test.options).to.deep.equal({ ss: 60, 'c:v': 'h264' });
      });

      it('removes third argument of three', () => {
        const test = new FfmpegSettings();
        test.addOption('ss', 60);
        test.addOption('t', 50);
        test.addOption('c:v', 'h264');
        expect(test.options).to.deep.equal({ ss: 60, t: 50, 'c:v': 'h264' }, 'Failed to set initial values to be deleted');
        test.removeOption('c:v');
        expect(test.options).to.deep.equal({ ss: 60, t: 50 });
      });

      it('does nothing with no parameters', () => {
        const test = new FfmpegSettings();
        test.addOption();
        expect(test.options).to.deep.equal({});
      });
    });

    describe('hasOption()', () => {
      it('returns true for an argument with a hyphen', () => {
        const test = new FfmpegSettings();
        test.addOption('-vn');
        expect(test.hasOption('vn')).to.be.true;
      });

      it('returns true for an argument without a hyphen', () => {
        const test = new FfmpegSettings();
        test.addOption('vn');
        expect(test.hasOption('vn')).to.be.true;
      });

      it('returns false for an argument with a hyphen', () => {
        const test = new FfmpegSettings();
        test.addOption('-vn');
        expect(test.hasOption('-an')).to.be.false;
      });

      it('returns false for an argument without a hyphen', () => {
        const test = new FfmpegSettings();
        test.addOption('vn');
        expect(test.hasOption('an')).to.be.false;
      });

      it('returns false when called without an argument', () => {
        const test = new FfmpegSettings(inputQuicktimePath);
        test.addOption('vn');
        expect(test.hasOption()).to.be.false;
      });
    });

    describe('removeOption()', () => {
      it('removes an argument', () => {
        const test = new FfmpegSettings();
        test.addOption('vn');
        test.addOption('an');
        expect(test.options, 'Test setup failed: options to remove not added').to.deep.equal({ vn: null, an: null });
        test.removeOption('an');
        expect(test.options).to.deep.equal({ vn: null });
      });

      it('does nothing when called with no arguments', () => {
        const test = new FfmpegSettings();
        test.addOption('vn');
        test.addOption('an');
        test.removeOption();
        expect(test.options).to.deep.equal({ vn: null, an: null });
      });
    });

    describe('metadata()', () => {
      it('adds a key and value argument', () => {
        const test = new FfmpegSettings();
        test.metadata('title', 'Fake Title');
        expect(test.customMetadata).to.deep.equal({ title: 'Fake Title' });
      });

      it('adds a single object with one key', () => {
        const test = new FfmpegSettings();
        test.metadata({ title: 'Fake Title' });
        expect(test.customMetadata).to.deep.equal({ title: 'Fake Title' });
      });

      it('adds a single object with multiple keys', () => {
        const test = new FfmpegSettings();
        test.metadata({
          title: 'Fake Title',
          artist: 'Fake Artist',
          album: 'Fake Album',
        });
        expect(test.customMetadata).to.deep.equal({
          title: 'Fake Title',
          artist: 'Fake Artist',
          album: 'Fake Album',
        });
      });
    });

    describe('getMetadataCommandLine()', () => {
      it('gets metadata without stream specifiers', () => {
        const test = new FfmpegSettings();
        test.customMetadata = {
          title: 'My Movie',
          artist: 'The Artist',
          album: 'Unknown Album',
        };
        expect(test.getMetadataCommandLine()).to.deep.equal([
          '-metadata', 'title="My Movie"',
          '-metadata', 'artist="The Artist"',
          '-metadata', 'album="Unknown Album"',
        ]);
      });

      it('gets metadata with a stream specifier', () => {
        const test = new FfmpegSettings();
        test.customMetadata = {
          title: 'My Movie',
          artist: 'The Artist',
          album: 'Unknown Album',
        };
        expect(test.getMetadataCommandLine('0:s:0')).to.deep.equal([
          '-metadata:0:s:0', 'title="My Movie"',
          '-metadata:0:s:0', 'artist="The Artist"',
          '-metadata:0:s:0', 'album="Unknown Album"',
        ]);
      });

      it('returns a empty array when no metadata set', () => {
        const test = new FfmpegSettings();
        expect(test.getMetadataCommandLine()).to.deep.equal([]);
        expect(test.getMetadataCommandLine('0:V')).to.deep.equal([]);
      });
    });

    describe('getOptionsCommandLine()', () => {
      it('gets options without stream specifiers', () => {
        const test = new FfmpegSettings();
        test.options = {
          vn: null,
          ss: 60,
          t: 10,
        };
        expect(test.getOptionsCommandLine()).to.deep.equal([
          '-vn',
          '-ss', 60,
          '-t', 10,
        ]);
      });

      it('gets options with a stream specifier', () => {
        const test = new FfmpegSettings();
        test.options = {
          vn: null,
          ss: 60,
          t: 10,
        };
        expect(test.getOptionsCommandLine('0:v')).to.deep.equal([
          '-vn',
          '-ss:0:v', 60,
          '-t:0:v', 10,
        ]);
      });

      it('returns a empty array when no options set', () => {
        const test = new FfmpegSettings();
        expect(test.getOptionsCommandLine()).to.deep.equal([]);
        expect(test.getOptionsCommandLine('0:V')).to.deep.equal([]);
      });
    });
  });

  describe('Integration Tests', () => {
    describe('getCommandLine()', () => {
      const test = new FfmpegSettings();
      test.customMetadata = {
        title: 'My Movie',
        artist: 'The Artist',
        album: 'Unknown Album',
      };
      test.options = {
        vn: null,
        ss: 60,
        t: 10,
      };
      expect(test.getCommandLine()).to.deep.equal([
        '-vn',
        '-ss', 60,
        '-t', 10,
        '-metadata', 'title="My Movie"',
        '-metadata', 'artist="The Artist"',
        '-metadata', 'album="Unknown Album"',
      ]);
    });
  });
});
