/* eslint camelcase: "off" */
const FfmpegSettings = require('../FfmpegSettings/FfmpegSettings');
const { FfmpegUnsupportedError } = require('../FfmpegErrors/FfmpegErrors');

const supportedDispositions = [
  0,
  'default',
  'dub',
  'original',
  'comment',
  'lyrics',
  'karaoke',
  'forced',
  'hearing_impaired',
  'visual_impaired',
  'clean_effects',
  'attached_pic',
  'captions',
  'descriptions',
  'dependent',
  'metadata',
];

const validHwaccels = [
  'none',
  'auto',
  'vdpau',
  'dxva2',
  'vaapi',
  'qsv',
];

module.exports = class FfmpegStream {
  constructor(streamData) {
    Object.assign(this, streamData);
    this.ffmpegSettings = new FfmpegSettings();
  }

  /**
   * Set the number of audio frames to output.
   *
   * @param {number} number
   */
  aframes(number) {
    this.ffmpegSettings.addOption('-frames:a', number);
    return this;
  }

  /**
   * Set the video display aspect ratio specified by aspect. Aspect can be a floating point number
   * string, or a string of the form num:den, where num and den are the numerator and denominator of
   * the aspect ratio. For example "4:3", "16:9", "1.3333", and "1.7777" are valid argument values.
   *
   * If used together with -vcodec copy, it will affect the aspect ratio stored at container level,
   * but not the aspect ratio stored in encoded frames, if it exists.
   *
   * @param {string|number} aspect
   */
  aspect(aspect) {
    this.ffmpegSettings.addOption('-aspect', aspect);
    return this;
  }

  /**
   * When doing stream copy, copy also non-key frames found at the beginning.
   */
  copyinkf() {
    this.ffmpegSettings.addOption('-copyinkf');
    return this;
  }

  /**
   * Intra_dc_precision
   */
  dc(precision) {
    this.ffmpegSettings.addOption('-dc', precision);
    return this;
  }

  /**
   *  Sets the disposition for a stream. This option overrides the disposition copied from the
   *  input stream. It is also possible to delete the disposition by setting it to 0. This option
   *  overrides the disposition copied from the input stream. It is also possible to delete the
   *  disposition by setting it to 0.
   *
   *  The following dispositions are recognized:
   *  default dub original comment lyrics karaoke forced hearing_impaired visual_impaired
   *  clean_effects attached_pic captions descriptions dependent metadata
   *
   * @param {string|0} value The disposition, or 0 to unset the option
   * @returns this
   */
  disposition(value) {
    if (supportedDispositions.includes(value)) {
      this.ffmpegSettings.addOption('-disposition', value);
      return this;
    }

    throw new FfmpegUnsupportedError(`${value} is not a supported disposition`);
  }

  /**
   * Create the filtergraph specified by filtergraph and use it to filter the stream.
   * filtergraph is a description of the filtergraph to apply to the stream, and must have a single
   * input and a single output of the same type of the stream. In the filtergraph, the input is
   * associated to the label in, and the output to the label out. See the ffmpeg-filters manual for
   * more information about the filtergraph syntax.
   *
   * See the filter_complex() method if you want to create filtergraphs with multiple inputs and/or
   * outputs.
   * @param filtergraph
   */
  // TODO: decide on shape of filtergraph parameter
  filter(filtergraph) {
    this.ffmpegSettings.addOption('-filter', filtergraph);
    return this;
  }

  /**
   * This option is similar to -filter, the only difference is that its argument is the name of the
   * file from which a filtergraph description is to be read.
   *
   * @param {string} filename the file to read the filter graph from
  */
  filter_script(filename) {
    this.ffmpegSettings.addOption('-filter_script', filename);
    return this;
  }

  // TODO
  force_key_frames() {
    return this;
  }

  /**
   * Stop writing to the stream after framecount frames.
   * @param {number} framecount
   */
  frames(framecount) {
    this.ffmpegSettings.addOption('-frames', framecount);
    return this;
  }

  /**
   * Use hardware acceleration to decode the matching stream(s). The allowed values of hwaccel are:
   * none auto vdpau dxva2 vaapi qsv
   *
   * This option has no effect if the selected hwaccel is not available or not supported by the
   * chosen decoder.
   *
   * Note that most acceleration methods are intended for playback and will not be faster than
   * software decoding on modern CPUs. Additionally, ffmpeg will usually need to copy the decoded
   * frames from the GPU memory into the system memory, resulting in further performance loss. This
   * option is thus mainly useful for testing.
   */
  hwaccel(name) {
    if (validHwaccels.includes(name)) {
      this.ffmpegSettings.addOption('-hwaccel', name);
      return this;
    }

    throw new FfmpegUnsupportedError(`${name} is not a supported hwaccel`);
  }

  /**
   * Set a metadata key/value pair. This option overrides metadata set with It is also possible to
   * delete metadata by using an empty value.
   *
   * @param {string|object} key Metadata field to be set, or an object of key-value pairs to set
   * @param {string} value Metadata value
   */
  metadata(key, value) {
    this.ffmpegSettings.metadata(key, value);
    return this;
  }

  /**
   * Specify the preset for matching stream.
   *
   * @param {string}
   */
  pre(preset_name) {
    this.ffmpegSettings.addOption('-pre', preset_name);
    return this;
  }

  /**
   * Alias for {@link qscale}
   */
  q(q) {
    return this.qscale(q);
  }

  /**
   * Use fixed quality scale (VBR). The meaning of q/qscale is codec-dependent. If qscale is used
   * without a stream_specifier then it applies only to the video stream, this is to maintain
   * compatibility with previous behavior and as specifying the same codec specific value to 2
   * different codecs that is audio and video generally is not what is intended when no
   * stream_specifier is used.
   * @param {number} q
   */
  qscale(q) {
    this.ffmpegSettings.addOption('-qscale', q);
    return this;
  }

  /**
   * Select the pass number (1 or 2). It is used to do two-pass video encoding. The statistics of
   * the video are recorded in the first pass into a log file (see also the option -passlogfile),
   * and in the second pass that log file is used to generate the video at the exact requested
   * bitrate. On pass 1, you may just deactivate audio and set output to null.
   *
   * @param {number} n
   */
  pass(n) {
    this.ffmpegSettings.addOption('-pass', n);
    return this;
  }

  /**
   * Set two-pass log file name prefix to prefix, the default file name prefix is “ffmpeg2pass”.
   * The complete file name will be PREFIX-N.log, where N is a number specific to the output stream
   */
  passlogfile(prefix) {
    this.ffmpegSettings.addOption('-prefix', prefix);
    return this;
  }

  /**
   * Set pixel format. Use -pix_fmts to show all the supported pixel formats. If the selected pixel
   * format can not be selected, ffmpeg will print a warning and select the best pixel format
   * supported by the encoder. If pix_fmt is prefixed by a +, ffmpeg will exit with an error if the
   * requested pixel format can not be selected, and automatic conversions inside filtergraphs are
   * disabled. If pix_fmt is a single +, ffmpeg selects the same pixel format as the input (or graph
   * output) and automatic conversions are disabled.
   *
   * @param format
   */
  pix_fmt(format) {
    this.ffmpegSettings.addOption('-pix_fmt', format);
    return this;
  }

  /**
   * Rate control override for specific intervals, formatted as "int,int,int" list separated with
   * slashes. Two first values are the beginning and end frame numbers, last one is quantizer to use
   * if positive, or quality factor if negative.
   *
   * @param {string}
   */
  rc_override(override) {
    this.ffmpegSettings.addOption('-rc_override', override);
    return this;
  }

  /**
   * top=1/bottom=0/auto=-1 field first
   */
  top(n) {
    this.ffmpegSettings.addOption('top', n);
    return this;
  }

  addOption(key, value) {
    this.ffmpegSettings.addOption(key, value);
    return this;
  }

  removeOption(key) {
    this.ffmpegSettings.removeOption(key);
    return this;
  }

  hasOption(key) {
    return this.ffmpegSettings.hasOption(key);
  }

  /**
   * Returns an object containing this instance's properties. If stringify is true, returns a JSON
   * string
   *
   * @param {boolean} stringify Returns a JSON string if true
   */
  getParsedData(stringify = false) {
    const json = Object.assign({}, this);
    delete json.ffmpegSettings;

    if (stringify) {
      return JSON.stringify(json);
    }

    return json;
  }

  getCommandLine(fileSpecifier = 0) {
    return this.ffmpegSettings.getCommandLine(`${fileSpecifier}:${this.index}`);
  }
};
