/* eslint camelcase: "off" */
const FfmpegStreamVideo = require('../FfmpegStreamVideo/FfmpegStreamVideo');
const { FfmpegUnsupportedError } = require('../FfmpegErrors/FfmpegErrors');

const supportedHwaccels = [
  'none',
  'auto',
  'vdpau',
  'dxva2',
  'vaapi',
  'qsv',
];

module.exports = class FfmpegStreamVideoInput extends FfmpegStreamVideo {
  constructor(streamData) {
    super();
    Object.assign(this, JSON.parse(JSON.stringify(streamData))); // deep copy
  }

  /**
   * Use hardware acceleration to decode the matching stream(s). The allowed values of hwaccel are:
   * none auto vdpau dxva2 vaapi qsv
   *
   * This option has no effect if the selected hwaccel is not available or not supported by the
   * chosen decoder.
   *
   * Note that most acceleration methods are intended for playback and will not be faster than
   * software decoding on modern CPUs. Additionally, ffmpeg will usually need to copy the decoded
   * frames from the GPU memory into the system memory, resulting in further performance loss. This
   * option is thus mainly useful for testing.
   */
  hwaccel(name) {
    if (supportedHwaccels.includes(name)) {
      return this.addOption('-hwaccel', name);
    }

    throw new FfmpegUnsupportedError(`${name} is not a supported hwaccel`);
  }

  /**
   * Set pixel format. Use -pix_fmts to show all the supported pixel formats. If the selected pixel
   * format can not be selected, ffmpeg will print a warning and select the best pixel format
   * supported by the encoder. If pix_fmt is prefixed by a +, ffmpeg will exit with an error if the
   * requested pixel format can not be selected, and automatic conversions inside filtergraphs are
   * disabled. If pix_fmt is a single +, ffmpeg selects the same pixel format as the input (or graph
   * output) and automatic conversions are disabled.
   *
   * @param format
   */
  pix_fmt(format) {
    return this.addOption('-pix_fmt', format);
  }

  /**
   * Set frame rate (Hz value, fraction or abbreviation).
   * As an input option, ignore any timestamps stored in the file and instead generate timestamps
   * assuming constant frame rate fps. This is not the same as the -framerate option used for some
   * input formats like image2 or v4l2 (it used to be the same in older versions of FFmpeg). If in
   * doubt use -framerate instead of the input option -r.
   *
   * @param {number} r
   */
  r(fps) {
    return this.addOption('-r', fps);
  }

  /**
   * Set frame size.
   * As an input option, this is a shortcut for the video_size private option, recognized by some
   * demuxers for which the frame size is either not stored in the file or is configurable – e.g.
   * raw video or video grabbers.
   *
   * @param {string} size In the format wxh
   */
  s(size) {
    return this.addOption('-s', size);
  }
};
