/* eslint camelcase: "off" */
const FfmpegStreamVideo = require('../FfmpegStreamVideo/FfmpegStreamVideo');
const { FfmpegUnsupportedError } = require('../FfmpegErrors/FfmpegErrors');

const supportedDispositions = [
  0,
  'default',
  'dub',
  'original',
  'comment',
  'lyrics',
  'karaoke',
  'forced',
  'hearing_impaired',
  'visual_impaired',
  'clean_effects',
  'attached_pic',
  'captions',
  'descriptions',
  'dependent',
  'metadata',
];

module.exports = class FfmpegStreamVideoOutput extends FfmpegStreamVideo {
  /**
   * Set the video display aspect ratio specified by aspect. Aspect can be a floating point number
   * string, or a string of the form num:den, where num and den are the numerator and denominator of
   * the aspect ratio. For example "4:3", "16:9", "1.3333", and "1.7777" are valid argument values.
   *
   * If used together with -vcodec copy, it will affect the aspect ratio stored at container level,
   * but not the aspect ratio stored in encoded frames, if it exists.
   *
   * @param {string|number} aspect
   */
  aspect(aspect) {
    return this.addOption('-aspect', aspect);
  }

  /**
   * When doing stream copy, copy also non-key frames found at the beginning.
   */
  copyinkf() {
    return this.addOption('-copyinkf');
  }

  /**
   * Intra_dc_precision
   */
  dc(precision) {
    return this.addOption('-dc', precision);
  }

  /**
   *  Sets the disposition for a stream. This option overrides the disposition copied from the
   *  input stream. It is also possible to delete the disposition by setting it to 0. This option
   *  overrides the disposition copied from the input stream. It is also possible to delete the
   *  disposition by setting it to 0.
   *
   *  The following dispositions are recognized:
   *  default dub original comment lyrics karaoke forced hearing_impaired visual_impaired
   *  clean_effects attached_pic captions descriptions dependent metadata
   *
   * @param {string|0} value The disposition, or 0 to unset the option
   * @returns this
   */
  disposition(value) {
    if (supportedDispositions.includes(value)) {
      return this.addOption('-disposition', value);
    }

    throw new FfmpegUnsupportedError(`${value} is not a supported disposition`);
  }

  /**
   * Create the filtergraph specified by filtergraph and use it to filter the stream.
   * filtergraph is a description of the filtergraph to apply to the stream, and must have a single
   * input and a single output of the same type of the stream. In the filtergraph, the input is
   * associated to the label in, and the output to the label out. See the ffmpeg-filters manual for
   * more information about the filtergraph syntax.
   *
   * See the filter_complex() method if you want to create filtergraphs with multiple inputs and/or
   * outputs.
   * @param filtergraph
   */
  // TODO: decide on shape of filtergraph parameter
  filter(filtergraph) {
    return this.addOption('-filter', filtergraph);
  }

  /**
   * This option is similar to -filter, the only difference is that its argument is the name of the
   * file from which a filtergraph description is to be read.
   *
   * @param {string} filename the file to read the filter graph from
  */
  filter_script(filename) {
    return this.addOption('-filter_script', filename);
  }

  // TODO
  force_key_frames() {
    return this;
  }

  /**
   * Stop writing to the stream after framecount frames.
   * @param {number} framecount
   */
  frames(framecount) {
    return this.addOption('-frames', framecount);
  }

  /**
   * Set a metadata key/value pair. This option overrides metadata set with It is also possible to
   * delete metadata by using an empty value.
   *
   * @param {string|object} key Metadata field to be set, or an object of key-value pairs to set
   * @param {string} value Metadata value
   */
  metadata(key, value) {
    this.ffmpegSettings.metadata(key, value);
  }

  /**
   * Specify the preset for matching stream.
   *
   * @param {string}
   */
  pre(preset_name) {
    return this.addOption('-pre', preset_name);
  }

  /**
   * Alias for {@link qscale}
   */
  q(q) {
    return this.qscale(q);
  }

  /**
   * Use fixed quality scale (VBR). The meaning of q/qscale is codec-dependent.
   * @param {number} q
   */
  qscale(q) {
    return this.addOption('-qscale', q);
  }

  /**
   * Select the pass number (1 or 2). It is used to do two-pass video encoding. The statistics of
   * the video are recorded in the first pass into a log file (see also the option -passlogfile),
   * and in the second pass that log file is used to generate the video at the exact requested
   * bitrate. On pass 1, you may just deactivate audio and set output to null.
   *
   * @param {number} n
   */
  pass(n) {
    return this.addOption('-pass', n);
  }

  /**
   * Set two-pass log file name prefix to prefix, the default file name prefix is “ffmpeg2pass”.
   * The complete file name will be PREFIX-N.log, where N is a number specific to the output stream
   */
  passlogfile(prefix) {
    return this.addOption('-prefix', prefix);
  }

  /**
   * Set frame rate (Hz value, fraction or abbreviation).
   *
   * Duplicate or drop input frames to achieve constant output frame rate fps.
   *
   * @param {number} r
   */
  r(fps) {
    return this.addOption('-r', fps);
  }

  /**
   * Rate control override for specific intervals, formatted as "int,int,int" list separated with
   * slashes. Two first values are the beginning and end frame numbers, last one is quantizer to use
   * if positive, or quality factor if negative.
   *
   * @param {string}
   */
  rc_override(override) {
    return this.addOption('-rc_override', override);
  }

  /**
   * Set frame size.
   *
   * This inserts the scale video filter to the end of the corresponding filtergraph. Please use the
   * scale filter directly to insert it at the beginning or some other place.
   *
   * @param {string} size In the format wxh
   */
  s(size) {
    return this.addOption('-s', size);
  }

  /**
   * top=1/bottom=0/auto=-1 field first
   */
  top(n) {
    return this.addOption('top', n);
  }
};
